;;;; coops-extras.scm  -*- Scheme -*-
;;;; Kon Lovett, Jul '18
;;;; Kon Lovett, Jun '17
;;;; Kon Lovett, Aug '10

(module coops-extras

(;export
  slot@
  make-copy)

(import scheme
  (chicken base)
  (chicken type)
  (only (srfi 1) cons* remove)
  (only coops-introspection check-instance)
  coops)

;;;

(define-type coops-class *)
(define-type coops-instance *)

(define-type plist list)

;;;coops-introspection

(: *class-slots (coops-class --> list))
;
(define (*class-slots class)
  (slot-value class 'slots) )

;;; Helpers

;;

;memq is not specific enough
(: plist-key? (symbol plist --> boolean))
;
(define (plist-key? key ls)
  ;search plist for key name
  (let loop ((ls ls))
    (and
      (not (null? ls))
      (or
        (eq? key (car ls))
        (loop (cddr ls)) ) ) ) )

(: plist-cons (symbol * plist --> plist))
;
(define (plist-cons key val ls) (cons* key val ls))

;;

;memq is not specific enough
(: initslot? (symbol list --> boolean))
;
(define (initslot? slot initforms) (plist-key? slot initforms))

(: slot-values (coops-instance (list-of symbol) #!optional list --> list))
;
(define (slot-values x slots #!optional (base '()))
  (foldl
    (lambda (ls slot)
      (if (slot-initialized? x slot)  ;per Jun 19, '17 email from Sandra Snan
        (plist-cons slot (slot-value x slot) ls)
        ls ) )
    base
    slots) )

(: shadowed-initforms (coops-instance list #!optional coops-class --> list))
;
(define (shadowed-initforms obj initforms #!optional (class (class-of obj)))
  (slot-values
    obj
    (remove (cut initslot? <> initforms) (*class-slots class))
    initforms) )

;;; Extras

;;

;sub-instance slot reference
(define-syntax slot@
  (syntax-rules (=)
    ;
    ((slot@ ?obj)
      ?obj )
    ;
    ((slot@ ?obj ?slot = ?valu)
      (set! (slot-value ?obj '?slot) ?valu) )
    ;
    ((slot@ ?obj ?slot ?rest0 ...)
      (slot@ (slot-value ?obj '?slot) ?rest0 ...) ) ) )

;;

;use w/ <primitive-object> is very suspect
(: make-copy (coops-instance #!rest --> *))
;
(define (make-copy obj . initforms)
  (let ((class (class-of (check-instance 'make-copy obj))))
    (apply make class (shadowed-initforms obj initforms class)) ) )

) ;coops-extras
